import os
import argparse

import yaml
import json


def is_file_type(filename):
    if not os.path.isfile(os.path.expanduser(filename)):
        raise argparse.ArgumentTypeError(f'path {filename} is not a file')
    return filename


def is_not_file_type(filename):
    if os.path.isfile(os.path.expanduser(filename)):
        raise argparse.ArgumentTypeError(f'path {filename} is an existing file')
    return filename


def load_filename(filename):
    if filename.endswith('.yaml') or filename.endswith('.yml'):
        return yaml.load(open(filename))
    elif filename.endswith('.json'):
        return json.load(open(filename))
    else:
        raise argparse.ArgumentError('only json and yaml files supported')



import subprocess


class DFTReader:
    @property
    def forces(self):
        """ return numpy.ndarray X x 3 in units [eV/Angstom] """
        raise NotImplementedError()

    @property
    def stress(self):
        """ return numpy.ndarray 3 x 3 in units [GPa] """
        raise NotImplementedError()

    @property
    def energy(self):
        """ return float in units [eV] """
        raise NotImplementedError()

    @property
    def structure(self):
        """ returns pymatgen.Structure. Not units for pymatgen are Angstroms """
        raise NotImplementedError()


class MDReader:
    def __init__(self, forces, stress, energy, structure):
        self._forces = forces
        self._stress = stress
        self._energy = energy
        self._structure = structure

    @property
    def forces(self):
        """ return numpy.ndarray X x 3 in units [eV/Angstom] """
        return self._forces

    @property
    def stress(self):
        """ return numpy.ndarray 3 x 3 in units [GPa] """
        return self._stress

    @property
    def energy(self):
        """ return float in units [eV] """
        return self._energy

    @property
    def structure(self):
        """ returns pymatgen.Structure. Not units for pymatgen are Angstroms """
        return self._structure


class DFTFITCalculator:
    """DFTFIT interface. Should have a simple api. Only `__init__`
    interface may change.

    """
    def __init__(self, structures):
        raise NotImplementedError()

    async def create(self):
        raise NotImplementedError()

    async def submit(self, potential, properties=None):
        properties = properties or {'stress', 'energy', 'forces'}
        raise NotImplementedError()


class MDCalculator:
    async def submit(self, structure, potential):
        raise NotImplementedError()



import pytest
import numpy as np

from dftfit.io.mattoolkit import MTKReader

# I don't really have plans to make mattoolkit available in the future
# @pytest.mark.mattoolkit
# def test_mattoolkit():
#     mtk = MTKReader(2) # super relaxation
#     assert np.all(np.isclose(mtk.forces, np.zeros((8, 3))))
#     assert np.all(np.isclose(mtk.stress, np.eye(3) * -36.97713))
#     assert np.isclose(mtk.energy, -47.47590796)
#     structure = mtk.structure
#     assert len(structure) == 8
#     assert set(s.symbol for s in structure.species) == {'Mg', 'O'}



from pymatgen.core import Specie, Element


def element_type_to_symbol(specie):
    if isinstance(specie, Specie):
        return specie.element.symbol
    elif isinstance(specie, Element):
        return specie.symbol
    elif isinstance(specie, str):
        return specie
    else:
        raise TypeError('Unsure how to convert to element symbol')



from marshmallow import fields, validate, ValidationError

from .base import BaseSchema


class KeyValueField(fields.Field):
    def _deserialize(self, value, attr, data):
        if value is None:
            return None
        if not isinstance(value, dict):
            self.fail('invalid must be key value dict')
        for key in value:
            if not isinstance(key, str) or not isinstance(value[key], str):
                self.fail('key value pairs must be strings')
        return value

    def _serialize(self, value, attr, data):
        return value


class MetadataSchema(BaseSchema):
    name = fields.String()
    labels = KeyValueField()


class ConfigurationSchema(BaseSchema):
    version = fields.String(required=True, validate=validate.Equal('v1'))
    kind = fields.String(required=True, validate=validate.Equal('Configuration'))
    metadata = fields.Nested(MetadataSchema)
    spec = fields.Dict()



from .base import MDReader


class MinimalMDReader(MDReader):
    def __init__(self, forces, stress, energy, structure):
        self._forces = forces
        self._stress = stress
        self._energy = energy
        self._structure = structure

    @property
    def forces(self):
        return self._forces

    @property
    def stress(self):
        return self._stress

    @property
    def energy(self):
        return self._energy

    @property
    def structure(self):
        return self._structure



from pathlib import Path

import numpy as np
from pymatgen.io.vasp import Vasprun

from .base import DFTReader


class VaspReader(DFTReader):
    def __init__(self, directory, vasprun_filename='vasprun.xml'):
        self.directory = Path(directory)
        self.vasprun_filename = vasprun_filename
        if not self.directory.is_dir():
            raise ValueError('path %s must exist and be directory' % self.directory)
        self._parse()

    def _parse(self):
        vasprun_path = list(self.directory.glob('**/%s' % self.vasprun_filename))
        if len(vasprun_path) > 1:
            raise ValueError('multiple vasprun.xml files found within directory')
        elif len(vasprun_path) == 0:
            raise ValueError('could not find vasprun.xml file within directory')

        vasprun = Vasprun(str(vasprun_path[0]))
        self._forces = np.array(vasprun.ionic_steps[-1]['forces'])
        self._stress = np.array(vasprun.ionic_steps[-1]['stress']) * 1e3 # kbar -> bar
        self._energy = float(vasprun.final_energy)
        self._structure = vasprun.final_structure

    @property
    def forces(self):
        return self._forces

    @property
    def stress(self):
        return self._stress

    @property
    def energy(self):
        return self._energy

    @property
    def structure(self):
        return self._structure



from marshmallow import Schema, ValidationError
from marshmallow.decorators import validates_schema


class BaseSchema(Schema):
    def __init__(self, strict=True, **kwargs):
        super(Schema, self).__init__(strict=strict, **kwargs)

    @validates_schema(pass_original=True, pass_many=False, skip_on_field_errors=True)
    def check_unknown_fields(self, data, original_data):
        def check_unknown(original_data_single):
            dump_only_keys = {key for key in self.fields if self.fields[key].dump_only}

            unknown_dump = set(dump_only_keys) & set(original_data_single)
            unknown_invalid = set(original_data_single) - set(self.fields.keys())
            unknown = unknown_dump | unknown_invalid
            if unknown:
                raise ValidationError('Unknown field', unknown)

        if isinstance(original_data, list):
            for original_data_single in original_data:
                check_unknown(original_data_single)
        else:
            check_unknown(original_data)



from collections import OrderedDict

import numpy as np
from pymatgen import Lattice, Specie, Structure
import pytest

from dftfit.potential import Potential

@pytest.mark.pymatgen_lammps
@pytest.mark.calculator
def test_lammps_reader():
    from pmg_lammps import LammpsInput, LammpsData, LammpsScript
    from dftfit.io.lammps import (
        LammpsReader, lammps_dftfit_set, modify_input_for_potential
    )

    base_directory = 'test_files/lammps/'
    lammps = LammpsReader(base_directory + 'mgo')
    assert np.all(np.isclose(lammps.forces, np.zeros((8, 3))))
    assert np.all(np.isclose(lammps.stress, np.eye(3) * 661.01685))
    assert np.isclose(lammps.energy, -85.34101)
    structure = lammps.structure
    assert len(structure) == 8
    assert set(s.symbol for s in structure.species) == {'Mg', 'O'}

@pytest.mark.pymatgen_lammps
@pytest.mark.calculator
def test_lammps_writer_buckingham():
    """Tests that given a structure and potential that a certain lammps
    input script is created.

    """
    from pmg_lammps import LammpsInput, LammpsData, LammpsScript
    from dftfit.io.lammps import (
        LammpsReader, lammps_dftfit_set, modify_input_for_potential
    )

    # Create structure
    supercell = (2, 2, 2)
    a = 4.1990858 # From evaluation of potential
    lattice = Lattice.from_parameters(a, a, a, 90, 90, 90)
    mg = Specie('Mg', 1.4)
    o = Specie('O', -1.4)
    atoms = [mg, o]
    sites = [[0, 0, 0], [0.5, 0.5, 0.5]]
    structure = Structure.from_spacegroup(225, lattice, atoms, sites)

    # Create Potential
    potential = Potential({
        'version': 'v1',
        'kind': 'Potential',
        'spec': {
            'charge': {
                'Mg': 1.4, 'O': -1.4
            },
            'kspace': {
                'type': 'pppm', 'tollerance': 1e-5
            },
            'pair': {
                'type': 'buckingham',
                'cutoff': 10.0,
                'parameters': [
                    {
                        'elements': ['Mg', 'Mg'],
                        'coefficients': [1309362.2766468062, 0.104, 0.0]
                    },
                    {
                        'elements': ['Mg', 'O'],
                        'coefficients': [9892.357, 0.20199, 0.0]
                    },
                    {
                        'elements': ['O', 'O'],
                        'coefficients': [2145.7345, 0.3, 30.2222]
                    }
                ]
            }
        }
    })

    lammps_input = LammpsInput(
        LammpsScript(lammps_dftfit_set),
        LammpsData.from_structure(structure))
    modify_input_for_potential(lammps_input, potential)

    output_script = [
        ('echo', 'both'),
        ('log', 'lammps.log'),
        ('units', 'metal'),
        ('dimension', 3),
        ('boundary', 'p p p'),
        ('atom_style', 'full'),
        ('read_data', 'initial.data'),
        ('kspace_style', 'pppm 0.000010'),
        ('pair_style', 'buck/coul/long 10.0'),
        ('pair_coeff', [
            '1 1 1309362.2766468062 0.104 0.0',
            '1 2 9892.357 0.20199 0.0',
            '2 2 2145.7345 0.3 30.2222'
        ]),
        ('set', [
            'type 1 charge 1.400000',
            'type 2 charge -1.400000',
        ]),
        ('dump', '1 all custom 1 mol.lammpstrj id type x y z fx fy fz'),
        ('dump_modify', '1 sort id'),
        ('thermo_modify', 'flush yes'),
        ('thermo_style', 'custom step etotal pxx pyy pzz pxy pxz pyz'),
        ('run', 0)
    ]
    assert OrderedDict(output_script) == lammps_input.lammps_script



import json
import hashlib
import datetime as dt
import numpy as np
import pandas as pd

from .potential import Potential


def _write_potential(dbm, potential):
    potential_str = json.dumps(potential.as_dict(with_parameters=False), sort_keys=True)
    potential_hash = hashlib.md5(potential_str.encode('utf-8')).hexdigest()

    result = dbm.connection.execute(
        'SELECT id FROM potential WHERE potential.hash = ?',
        (potential_hash,)).fetchone()
    if result:
        return result['id']

    cursor = dbm.connection.execute(
        'INSERT INTO potential (hash, schema) VALUES (?, ?)',
        (potential_hash, potential.as_dict(with_parameters=False))
    )
    return cursor.lastrowid


def _write_training(dbm, training):
    training_str = json.dumps(training.schema, sort_keys=True)
    training_hash = hashlib.md5(training_str.encode('utf-8')).hexdigest()

    result = dbm.connection.execute(
        'SELECT id FROM training WHERE training.hash = ?',
        (training_hash,)).fetchone()
    if result:
        return result['id']

    cursor = dbm.connection.execute(
        'INSERT INTO training (hash, schema) VALUES (?, ?)',
        (training_hash, training.schema)
    )
    return cursor.lastrowid


def _write_labels(dbm, run_id, labels):
    for key, value in labels.items():
        result = dbm.connection.execute('''
        SELECT id FROM label
        WHERE label.key = ? AND label.value = ?
        ''', (key, value)).fetchone()

        if result:
            label_id = result['id']
        else:
            cursor = dbm.connection.execute('''
            INSERT INTO label (key, value)
            VALUES (?, ?)
            ''', (key, value))
            label_id = cursor.lastrowid

        cursor = dbm.connection.execute('''
        INSERT INTO run_label (run_id, label_id)
        VALUES (?, ?)
        ''', (run_id, label_id))


def write_run_initial(dbm, potential, training, configuration):
    with dbm.connection:
        potential_id = _write_potential(dbm, potential)
        training_id = _write_training(dbm, training)
        cursor = dbm.connection.execute('''
        INSERT INTO run (name, potential_id, training_id, configuration, start_time, initial_parameters, indicies, bounds)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?)
        ''', (
            configuration.run_name,
            potential_id,
            training_id,
            configuration.schema,
            dt.datetime.utcnow(),
            potential.parameters.tolist(),
            potential.optimization_parameter_indicies.tolist(),
            potential.optimization_bounds.tolist()
        ))
        run_id = cursor.lastrowid
        _write_labels(dbm, run_id, configuration.run_labels)
        return potential_id, run_id


def write_run_final(dbm, run_id):
    with dbm.connection:
        cursor = dbm.connection.execute('''
        UPDATE run SET end_time = ?
        WHERE id = ? AND end_time IS NULL
        ''', (dt.datetime.utcnow(), run_id))


def write_evaluation(dbm, run_id, potential, result):
    with dbm.connection:
        errors = (
            result['parts']['forces'],
            result['parts']['stress'],
            result['parts']['energy'])
        if 'weights' in result:
            weights = (
                result['weights']['forces'],
                result['weights']['stress'],
                result['weights']['energy']
            )
        else:
            weights = (0, 0, 0)
        cursor = dbm.connection.execute('''
        INSERT INTO evaluation (run_id, parameters, sq_force_error, sq_stress_error, sq_energy_error, w_f, w_s, w_e)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?)
        ''', (run_id, potential.optimization_parameters.tolist(),
              *errors, *weights))


def filter_evaluations(dbm, potential=None, limit=10, condition='best', run_id=None, labels=None):
    """Select a subset of evaluations. Currently only works on single
    objective functions because "best" and "worst" are subjective in
    multiobjecive functions.

    Arguments:
     - condition (str): best, random, worst

    THIS IS THE REASON TO USE SQLALCHEMY QUERY BUILDER.... I HAVE LEARNED NOW
    """

    query = '''
    SELECT run.id
    FROM run {join_sql}
    WHERE {where_sql}
    '''

    join_statement = []
    where_statement = ['1=1']
    query_arguments = []
    if labels:
        join_statement.extend([
            'JOIN run_label ON run.id = run_label.run_id',
            'JOIN label ON run_label.label_id = label.id'
        ])
        select_labels = []
        for key, value in labels.items():
            if not isinstance(key, str) or not isinstance(value, str):
                raise ValueError(f'key: {key} values: {values} for label must be strings')
            select_labels.append('(label.key = ? AND label.value = ?)')
            query_arguments.extend([key, value])
        where_statement.append(' OR '.join(select_labels))
        query += f"""
        GROUP BY run.id
        HAVING count = {len(labels)}
        """

    if run_id is not None:
        if not isinstance(run_id, int):
            raise ValueError('run_id must be integer')
        where_statement.append('run.id = ?')
        query_arguments.append(run_id)

    if potential:
        potential_str = json.dumps(potential.as_dict(with_parameters=False), sort_keys=True)
        potential_hash = hashlib.md5(potential_str.encode('utf-8')).hexdigest()
        join_statement.append('JOIN potential ON run.potential_id = potential.id')
        where_statement.append('potential.hash = ?')
        query_arguments.append(potential_hash)

    # find all run ids that match selection
    query = query.format(join_sql=' '.join(join_statement),
                         where_sql=' AND '.join(where_statement))
    run_ids = [row['id'] for row in dbm.connection.execute(query, query_arguments)]

    if condition == 'best':
        order_sql = 'score ASC'
    elif condition == 'worst':
        order_sql = 'score DESC'
    elif condition == 'random':
        order_sql = 'RANDOM()'
    else:
        raise ValueError('condition ordering not supported')

    # pick subset of potentials
    query = f'''
    SELECT e.id, e.parameters, (e.w_f * e.sq_force_error + e.w_s * e.sq_stress_error + e.w_e * e.sq_energy_error) AS score FROM evaluation e
    WHERE {' OR '.join(['e.run_id = ?' for _ in run_ids])}
    ORDER BY {order_sql}
    LIMIT {limit}
    '''
    return dbm.connection.execute(query, run_ids)


def filter_potentials(dbm, potential=None, limit=10, condition='best', run_id=None, labels=None):
    results = []
    for row in filter_evaluations(dbm, potential, limit, condition, run_id, labels):
        results.append({'potential': select_potential_from_evaluation(dbm, row['id']), 'score': row['score']})
    return results


def select_potential_from_evaluation(dbm, evaluation_id):
    result = dbm.connection.execute('''
    SELECT potential.schema, parameters, initial_parameters, indicies, bounds
    FROM evaluation
        JOIN run ON run.id = evaluation.run_id
        JOIN potential ON potential.id = run.potential_id
    WHERE evaluation.id = ?
    ''', (evaluation_id,)).fetchone()
    if result is None:
        raise ValueError(f'evaluation_id {evaluation_id} does not exist')

    return Potential.from_run_evaluation(
        result['schema'],
        result['initial_parameters'],
        result['indicies'], result['parameters'], result['bounds'])


def copy_database_to_database(src_dbm, dest_dbm, only_unique=False):
    SELECT_RUNS = 'SELECT id FROM run'
    SELECT_RUN = 'SELECT id, name, potential_id, training_id, configuration, start_time, end_time, initial_parameters, indicies, bounds FROM run WHERE id = ?'
    SELECT_RUN_LABELS = 'SELECT label.key, label.value FROM run_label JOIN label ON run_label.label_id = label.id WHERE run_label.run_id = ?'
    SELECT_RUN_POTENTIAL = 'SELECT potential.hash, potential.schema FROM potential JOIN run ON run.potential_id = potential.id WHERE run.id = ?'
    SELECT_RUN_TRAINING = 'SELECT training.hash, training.schema FROM training JOIN run ON run.potential_id = training.id WHERE run.id = ?'
    SELECT_RUN_EVALUATION_COUNT = 'SELECT count(*) as num_evaluations FROM evaluation WHERE run_id = ?'
    SELECT_RUN_EVALUATION = '''
    SELECT parameters, sq_force_error, sq_stress_error, sq_energy_error, w_f, w_s, w_e FROM evaluation
    WHERE run_id = ? ORDER BY id LIMIT ? OFFSET ?
    '''

    UNIQUE_RUN_POTENTIAL = 'SELECT id, hash FROM potential WHERE potential.hash = ?'
    UNIQUE_RUN_TRAINING = 'SELECT id, hash FROM training WHERE training.hash = ?'
    UNIQUE_RUN = '''
    SELECT run.id FROM run
              JOIN potential ON potential.id = run.potential_id
              JOIN training ON training.id = run.training_id
    WHERE name = ? AND potential.hash = ? AND training.hash = ? AND configuration = ?
                   AND start_time = ? AND (end_time = ? OR end_time IS NULL AND ? is NULL)
                   AND initial_parameters = ? AND indicies = ? AND bounds = ?
    '''

    INSERT_RUN_POTENTIAL = 'INSERT INTO potential (hash, schema) VALUES (?, ?)'
    INSERT_RUN_TRAINING = 'INSERT INTO training (hash, schema) VALUES (?, ?)'
    INSERT_RUN = 'INSERT INTO run (name, potential_id, training_id, configuration, start_time, end_time, initial_parameters, indicies, bounds) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)'
    INSERT_RUN_EVALUATION = 'INSERT INTO evaluation (run_id, parameters, sq_force_error, sq_stress_error, sq_energy_error, w_f, w_s, w_e) VALUES (?, ?, ?, ?, ?, ?, ?, ?)'

    for run in src_dbm.connection.execute(SELECT_RUNS):
        # Potential
        query_result = src_dbm.connection.execute(SELECT_RUN_POTENTIAL, (run['id'],)).fetchone()
        potential_hash = query_result['hash']
        query = dest_dbm.connection.execute(UNIQUE_RUN_POTENTIAL, (potential_hash,)).fetchone()
        if query:
            potential_id = query['id']
        else:
            with dest_dbm.connection:
                cursor = dest_dbm.connection.execute(INSERT_RUN_POTENTIAL, (query_result['hash'], query_result['schema']))
                potential_id = cursor.lastrowid

        # Training
        query_result = src_dbm.connection.execute(SELECT_RUN_TRAINING, (run['id'],)).fetchone()
        training_hash = query_result['hash']
        query = dest_dbm.connection.execute(UNIQUE_RUN_TRAINING, (training_hash,)).fetchone()
        if query:
            training_id = query['id']
        else:
            with dest_dbm.connection:
                cursor = dest_dbm.connection.execute(INSERT_RUN_TRAINING, (query_result['hash'], query_result['schema']))
                training_id = cursor.lastrowid

        # Run
        query_result = src_dbm.connection.execute(SELECT_RUN, (run['id'],)).fetchone()
        query = dest_dbm.connection.execute(UNIQUE_RUN, (
            query_result['name'], potential_hash, training_hash, query_result['configuration'],
            query_result['start_time'], query_result['end_time'], query_result['end_time'],
            query_result['initial_parameters'], query_result['indicies'], query_result['bounds'])).fetchone()
        if query and only_unique:
            run_id = query['id']
        else:
            with dest_dbm.connection:
                cursor = dest_dbm.connection.execute(INSERT_RUN, (
                    query_result['name'], potential_id, training_id, query_result['configuration'],
                    query_result['start_time'], query_result['end_time'],
                    query_result['initial_parameters'], query_result['indicies'], query_result['bounds']))
                run_id = cursor.lastrowid

            # Evaluation
            num_evaluations = src_dbm.connection.execute(SELECT_RUN_EVALUATION_COUNT, (run['id'],)).fetchone()['num_evaluations']
            print('   adding run %d with %d evaluations' % (run['id'], num_evaluations))
            evaluation_limit = 1000
            for offset in range(0, num_evaluations, evaluation_limit):
                cursor = src_dbm.connection.execute(SELECT_RUN_EVALUATION, (run['id'], evaluation_limit, offset))
                evaluations = [(run_id, row['parameters'], row['sq_force_error'], row['sq_stress_error'], row['sq_energy_error'], row['w_f'], row['w_s'], row['w_e']) for row in cursor]
                with dest_dbm.connection:
                    dest_dbm.connection.executemany(INSERT_RUN_EVALUATION, evaluations)

            # labels
            labels = {row['key']: row['value'] for row in src_dbm.connection.execute(SELECT_RUN_LABELS, (run['id'],))}
            with dest_dbm.connection:
                _write_labels(dest_dbm, run_id, labels)


def list_runs(dbm):
    SELECT_RUNS = 'SELECT id FROM run'
    return [row['id'] for row in dbm.connection.execute(SELECT_RUNS)]


def list_evaluations(dbm, run_id):
    SELECT_EVALUATIONS = f'SELECT sq_force_error, sq_stress_error, sq_energy_error, w_f, w_s, w_e FROM evaluation WHERE run_id = {run_id}'
    df = pd.read_sql(SELECT_EVALUATIONS, dbm.connection)
    df['score'] = df['w_f'] * df['sq_force_error'] + df['w_s'] * df['sq_stress_error'] + df['w_e'] * df['sq_energy_error']
    return df


def run_summary(dbm, run_id):
    SELECT_RUN = 'SELECT id, name, potential_id, training_id, configuration, start_time, end_time, initial_parameters, indicies, bounds FROM run WHERE id = ?'

    SELECT_RUN_NUM_EVALUATIONS = 'SELECT count(*) FROM evaluation WHERE run_id = ?'
    SELECT_RUN_LAST_EVALUATIONS = '''
    SELECT (e.w_f * e.sq_force_error + e.w_s * e.sq_stress_error + e.w_e * e.sq_energy_error) AS score FROM evaluation e
    WHERE run_id = ?
    ORDER BY e.id DESC LIMIT 100
    '''

    run = dbm.connection.execute(SELECT_RUN, (run_id,)).fetchone()
    run_summary = {
        'algorithm': run['configuration']['spec']['algorithm']['name'],
        'initial_parameters': run['initial_parameters'],
    }
    num_evaluations = dbm.connection.execute(SELECT_RUN_NUM_EVALUATIONS, (run_id,)).fetchone()[0]
    run_summary.update({
        'steps': num_evaluations
    })
    last_scores = [row['score'] for row in dbm.connection.execute(SELECT_RUN_LAST_EVALUATIONS, (run_id,))]
    if last_scores:
        run_summary.update({
            'stats': {'mean': np.mean(last_scores), 'median': np.median(last_scores), 'min': np.min(last_scores)}
        })
        min_score = filter_evaluations(dbm, run_id=run_id, condition='best', limit=1).fetchone()
        run_summary.update({
            'final_parameters': min_score['parameters'],
            'min_score': min_score['score']
        })
    else:
        run_summary.update({
            'stats': {'mean': 0.0, 'median': 0.0, 'min': 0.0},
            'final_parameters': [],
            'min_score': 0.0
        })
    return run_summary



from .potential import Potential
from .training import Training
from .config import Configuration
from .optimize import Optimize

from .db_actions import write_run_initial, write_run_final


def dftfit(configuration_schema, potential_schema, training_schema):
    configuration = Configuration(configuration_schema)
    potential = Potential(potential_schema)
    training = Training(training_schema, **configuration.training_kwargs)

    try:
        potential_id, run_id = write_run_initial(configuration.dbm, potential, training, configuration)
        optimize = Optimize(
            dft_calculations=training.calculations,
            potential=potential,
            dbm=configuration.dbm,
            algorithm=configuration.algorithm,
            algorithm_kwargs=configuration.algorithm_kwargs,
            problem_kwargs=configuration.problem_kwargs,
            run_id=run_id
        )
        population = optimize.population(configuration.population, seed=configuration.seed)
        optimize.optimize(population, steps=configuration.steps, seed=configuration.seed)
    except KeyboardInterrupt:
        print(f'\nShutting down DFTFIT\nIf using database all completed evaluations are written')
    finally:
        write_run_final(configuration.dbm, run_id)
    return run_id



import sys

from marshmallow import fields, validate, ValidationError

from .base import BaseSchema
from ..data import element_symbols
from ..parameter import FloatParameter


class ParameterSchema(BaseSchema):
    initial = fields.Float(required=True)
    bounds = fields.List(fields.Float(), validate=validate.Length(equal=2), missing=(-sys.float_info.max, sys.float_info.max))


class FloatParameterField(fields.Field):
    def _deserialize(self, value, attr, data):
        try:
            value = float(value)
            return FloatParameter(value, fixed=True)
        except (ValueError, TypeError):
            schema_load, errors = ParameterSchema().load(value)
            return FloatParameter(**value)

    def _serialize(self, value, attr, data):
        if value.fixed and not value.computed:
            return float(value)
        else:
            return {'initial': float(value), 'bounds': value.bounds}

    def _validate(self, value):
        if value is None:
            return None
        try:
            super()._validate(float(value))
        except (ValueError, TypeError):
            schema_load, errors = ParameterSchema().load(value)
            if not (schema_load['bounds'][0] < schema_load['initial'] < schema_load['bounds'][1]):
                raise ValidationError('initial value must be between bounds', 'initial')
            super()._validate(schema_load['initial'])


class ConstraintSchema(BaseSchema):
    charge_balance = fields.String()


ChargesSchema = type('ChargeSchema', (BaseSchema,), {element: FloatParameterField() for element in element_symbols})


class KspaceSchema(BaseSchema):
    KSPACE_TYPES = {'ewald', 'pppm'}

    type = fields.String(required=True, validate=validate.OneOf(KSPACE_TYPES))
    tollerance = FloatParameterField(required=True, validate=validate.Range(min=1e-16))


class ParametersSchema(BaseSchema):
    elements = fields.List(fields.String(validate=validate.OneOf(element_symbols)), required=True)
    coefficients = fields.List(FloatParameterField())


class PairPotentialSchema(BaseSchema):
    PAIR_POTENTIALS = {'buckingham'}

    type = fields.String(required=True, validate=validate.OneOf(PAIR_POTENTIALS))
    cutoff = FloatParameterField(required=False, validate=validate.Range(min=1e-6))
    parameters = fields.Nested(ParametersSchema, required=True, many=True)


class PotentialSpecSchema(BaseSchema):
    constraint = fields.Nested(ConstraintSchema)
    charge = fields.Nested(ChargesSchema)
    kspace = fields.Nested(KspaceSchema)
    pair = fields.Nested(PairPotentialSchema)


class PotentialSchema(BaseSchema):
    version = fields.String(required=True, validate=validate.Equal('v1'))
    kind = fields.String(required=True, validate=validate.Equal('Potential'))
    spec = fields.Nested(PotentialSpecSchema)



""" Handles collection of trial and testing data.

Has a caching layer as to speed up future runs
"""
import json
import shelve
import os

import yaml

from .schema import TrainingSchema
from .io import MTKReader


class Training:
    def __init__(self, schema, cache_filename=None):
        schema_load, errors = TrainingSchema().load(schema)
        self.schema = schema_load
        self._gather_calculations(cache_filename=cache_filename)

    def _gather_calculations(self, cache_filename=None):
        self._calculations = []
        for calculation in self.schema['spec']:
            if calculation['type'] == 'mattoolkit':
                self._calculations.extend(self.download_mattoolkit_calculations(calculation['selector'], cache_filename=cache_filename))

    @property
    def calculations(self):
        return self._calculations

    def __iter__(self):
        return iter(self._calculations)

    def __len__(self):
        return len(self._calculations)

    def download_mattoolkit_calculations(self, selector, cache_filename=None):
        from mattoolkit.api import CalculationResourceList

        if cache_filename:
            cache_directory, filename = os.path.split(cache_filename)
            os.makedirs(cache_directory, exist_ok=True)
            key = f'mattoolkit.calculation.' + '.'.join(selector['labels'])
            with shelve.open(cache_filename) as cache:
                if key in cache:
                    calc_ids = cache[key]
                else:
                    calculations = CalculationResourceList()
                    calculations.get(params={'labels': selector['labels']})
                    calc_ids = [c.id for c in calculations.items]
                    cache[key] = calc_ids
        else:
            calculations = CalculationResourceList()
            calculations.get(params={'labels': selector['labels']})
            calc_ids = [c.id for c in calculations.items]
        return [MTKReader(calc_id, cache_filename=cache_filename) for calc_id in calc_ids]

    @classmethod
    def from_file(cls, filename, format=None, **kwargs):
        if format not in {'json', 'yaml'}:
            if filename.endswith('json'):
                format = 'json'
            elif filename.endswith('yaml') or filename.endswith('yml'):
                format = 'yaml'
            else:
                raise ValueError('unrecognized filetype from filename %s' % filename)

        if format == 'json':
            with open(filename) as f:
                return cls(json.load(f))
        elif format in {'yaml', 'yml'}:
            with open(filename) as f:
                return cls(yaml.load(f), **kwargs)

    def __str__(self):
        return json.dumps(self.schema, sort_keys=True, indent=4)



from marshmallow import fields, validate, ValidationError

from .base import BaseSchema
from .fields import PolyField


class MTKSelectorSchema(BaseSchema):
    labels = fields.List(fields.String(), required=True)


class MTKTrainingSetSchema(BaseSchema):
    type = fields.String(required=True, validate=validate.Equal('mattoolkit'))
    selector = fields.Nested(MTKSelectorSchema, required=True)


def training_property_schema_serialization_disambiguation(base_object, obj):
    type_to_schema = {
        'mattoolkit': MTKTrainingSetSchema,
    }
    try:
        return type_to_schema[obj.mode]()
    except KeyError:
        pass
    raise TypeError("Could not detect type did you specify a type?")


def trainging_property_schema_deserialization_disambiguation(object_dict, data):
    type_to_schema = {
        'mattoolkit': MTKTrainingSetSchema,
    }
    try:
        return type_to_schema[object_dict['type']]()
    except KeyError:
        pass
    raise TypeError("Could not detect type did you specify a type?")


class TrainingSchema(BaseSchema):
    version = fields.String(required=True, validate=validate.Equal('v1'))
    kind = fields.String(required=True, validate=validate.Equal('Training'))
    spec = PolyField(
        serialization_schema_selector=training_property_schema_serialization_disambiguation,
        deserialization_schema_selector=trainging_property_schema_deserialization_disambiguation,
        many=True
    )



from ..dftfit import dftfit
from .utils import load_filename, is_file_type


def add_subcommand_train(subparsers):
    parser = subparsers.add_parser('train', help='train potential model')
    parser.set_defaults(func=handle_subcommand_train)
    parser.add_argument('-t', '--training', help='training set filename in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('-c', '--config', help='configuration filename in yaml/json format', type=is_file_type)
    parser.add_argument('-s', '--set', help='set specific features of config files', nargs='+', action='append')
    parser.add_argument('-d', '--database', help='sqlite database filename')
    parser.add_argument('--steps', help='number of steps to preform for fitting', type=int)
    parser.add_argument('--population', help='population size to begin with', type=int)
    parser.add_argument('-l', '--loglevel', help='logging level', )


def handle_subcommand_train(args):
    training_schema = load_filename(args.training)
    potential_schema = load_filename(args.potential)
    configuration_schema = {}
    if args.config:
        configuration_schema = load_filename(args.config)
    if args.database:
        configuration_schema['spec']['database'] = args.database
    if args.loglevel:
        configuration_schema['spec']['logging'] = args.loglevel
    if args.steps:
        configuration_schema['spec']['steps'] = args.steps
    if args.population:
        configuration_schema['spec']['population'] = args.population
    if args.set:
        raise NotImplementedError('Hope to one day support setting specific arguments')

    dftfit(training_schema=training_schema,
           potential_schema=potential_schema,
           configuration_schema=configuration_schema)



def print_elastic_information(elastic):
    print('Stiffness Tensor')
    for row in elastic.voigt:
        print('{:+8.1f} {:+8.1f} {:+8.1f} {:+8.1f} {:+8.1f} {:+8.1f}\n'.format(*row))
    print('\n')
    print('Shear Modulus G_V', elastic.g_voigt)
    print('Shear Modulus G_R', elastic.g_reuss)
    print('Shear Modulus G_vrh', elastic.g_vrh)

    print('Bulk Modulus K_V', elastic.k_voigt)
    print('Bulk Modulus K_R', elastic.k_reuss)
    print('Bulk Modulus K_vrh', elastic.k_vrh)

    print('Elastic Anisotropy', elastic.universal_anisotropy)
    print('Poisons Ration', elastic.homogeneous_poisson)



from pathlib import Path
from xml.etree import ElementTree

import numpy as np
import pymatgen as pmg

from .base import DFTReader


class SiestaReader(DFTReader):
    def __init__(self, directory, output_filename='output.xml'):
        self.directory = Path(directory)
        self.output_filename = output_filename
        if not self.directory.is_dir():
            raise ValueError('path %s must exist and be directory' % self.directory)
        self._parse()

    def _parse(self, step=-1):
        # for now just take the last step
        self._namespaces = {'xml': 'http://www.xml-cml.org/schema'}
        root = ElementTree.parse(self.directory / self.output_filename).getroot()
        self._parse_structure(root, step)
        self._parse_energy(root, step)
        self._parse_stresses(root, step)
        self._parse_forces(root, step)

    def _parse_structure(self, root, step=-1):
        XML_ATOMARRAY = './/xml:module[@dictRef="MD"]/xml:molecule[1]/xml:atomArray'
        xml_structures = root.findall(XML_ATOMARRAY, namespaces=self._namespaces)
        coordinates = []
        symbols = []
        for atom in xml_structures[step]:
            symbols.append(atom.attrib["elementType"])
            coordinates.append([atom.attrib["x3"], atom.attrib["y3"], atom.attrib["z3"]])
        coordinates = np.array(coordinates).reshape(-1, 3).astype(float)

        XML_LATTICE = './/xml:crystal[@title="Lattice Parameters"]'
        xml_lattices = root.findall(XML_LATTICE, namespaces=self._namespaces)
        xml_lattice = xml_lattices[step]
        XML_LATTICE_LENGTHS = './xml:cellParameter[@parameterType="length"]'
        XML_LATTICE_ANGLES = './xml:cellParameter[@parameterType="angle"]'
        lattice_lengths = [float(_) for _ in xml_lattice.find(XML_LATTICE_LENGTHS, namespaces=self._namespaces).text.split()]
        lattice_angles = [float(_) for _ in xml_lattice.find(XML_LATTICE_ANGLES, namespaces=self._namespaces).text.split()]

        lattice = pmg.Lattice.from_parameters(*lattice_lengths, *lattice_angles)
        self._structure = pmg.Structure(lattice, symbols, coordinates, coords_are_cartesian=True)

    def _parse_energy(self, root, step=-1):
        XML_ENERGIES = './/xml:propertyList[@title="Final KS Energy"]/xml:property/xml:scalar'
        energies = root.findall(XML_ENERGIES, namespaces=self._namespaces)
        self._energy = float(energies[step].text)

    def _parse_stresses(self, root, step=-1):
        eVA32GPa = 160.21766208 # http://greif.geo.berkeley.edu/~driver/conversions.html
        XML_STRESS = './/xml:property[@title="Total Stress"]/xml:matrix'
        xml_stresses = root.findall(XML_STRESS, namespaces=self._namespaces)
        self._stress = np.array([float(_) for _ in xml_stresses[step].text.split()]).reshape(3, 3) * eVA32GPa

    def _parse_forces(self, root, step=-1):
        XML_FORCES = './/xml:propertyList[@title="Forces"]/xml:property/xml:matrix'
        xml_forces = root.findall(XML_FORCES, namespaces=self._namespaces)
        self._forces = np.array([float(_) for _ in xml_forces[step].text.split()]).reshape(-1, 3)

    @property
    def forces(self):
        return self._forces

    @property
    def stress(self):
        return self._stress

    @property
    def energy(self):
        return self._energy

    @property
    def structure(self):
        return self._structure



import pytest

import numpy as np

from dftfit.io.siesta import SiestaReader


@pytest.mark.siesta
def test_siesta_reader():
    filename = 'd3_o_20ev.xml'
    directory = 'test_files/siesta'
    calculation = SiestaReader(directory, filename)
    assert np.all(np.isclose(calculation.energy, -104742.133616))

    first_row_forces = np.array([-5.014637520260e-1, -4.224890317363e-1, -1.420257672235e-1])
    assert np.all(np.isclose(calculation.forces[0], first_row_forces))

    eVA32GPa = 160.21766208 # http://greif.geo.berkeley.edu/~driver/conversions.html
    stresses = np.array([
        [-2.765925809224e-3, -3.009750267323e-5, 1.171322722617e-4],
        [-2.908082457191e-5, -3.180769833963e-3, -1.264574964357e-4],
        [1.174490293284e-4, -1.258277686581e-4, -1.767562635815e-3]
    ]) * eVA32GPa
    assert np.all(np.isclose(calculation.stress, stresses))



import math

import lammps
import numpy as np

from .base import DFTFITCalculator, MDReader


class LammpsCythonDFTFITCalculator(DFTFITCalculator):
    """This is not a general purpose lammps calculator. Only for dftfit
    evaluations. For now there are not plans to generalize it.
    """
    def __init__(self, structures, num_workers=1):
        if num_workers != 1:
            raise NotImplemented('For now limited to one worker')
        self.structures = structures
        self.lammps_systems = []

    def _initialize_lammps(self, structure):
        lmp = lammps.Lammps(units='metal', style='full', args=[
            '-log', 'none', '-screen', 'none'
        ])
        lmp.command('atom_modify map yes')
        elements = [s for s in set(structure.species)]
        atom_types = np.array([elements.index(atom.specie)+1 for atom in structure], dtype=np.intc)
        lmp.box.from_lattice_const(
            len(elements),
            np.array(structure.lattice.abc),
            np.array(structure.lattice.angles) * (math.pi/180))
        for element, atom_type in zip(elements, lmp.system.atom_types):
            atom_type.mass = element.atomic_mass
        lmp.system.create_atoms(atom_types, structure.cart_coords+1e-8)
        lmp.thermo.add('my_ke', 'ke', 'all')
        return lmp, elements

    async def create(self):
        for structure in self.structures:
            self.lammps_systems.append(self._initialize_lammps(structure))

    def _apply_potential(self, lmp, elements, potential):
        """Apply specific potential

        Wish that this was more generic. But I have found that simpler
        implementation is better for now.
        """
        # buckingham + charge
        spec = potential.schema['spec']
        if 'charge' in spec and 'kspace' in spec and 'pair' in spec and \
           spec['pair']['type'] == 'buckingham':
            self._apply_buckingham_charge(lmp, elements, potential)

    def _apply_buckingham_charge(self, lmp, elements, potential):
        element_map = {e.symbol: i for i, e in enumerate(elements, start=1)}
        spec = potential.schema['spec']
        lmp.command('kspace_style %s %f' % (
            spec['kspace']['type'], spec['kspace']['tollerance']))
        lmp.command('pair_style buck/coul/long %f' % spec['pair']['cutoff'])
        for parameter in spec['pair']['parameters']:
            ij = sorted([element_map[parameter['elements'][0]],
                         element_map[parameter['elements'][1]]])
            lmp.command('pair_coeff %d %d %s' % (
                ij[0], ij[1],
                ' '.join([str(float(coeff)) for coeff in parameter['coefficients']])))
        for element, charge in spec['charge'].items():
            lmp.command('set type %d charge %f' % (element_map[element], float(charge)))

    async def submit(self, potential, properties=None):
        properties = properties or {'stress', 'energy', 'forces'}
        results = []
        for (lmp, elements), structure in zip(self.lammps_systems, self.structures):
            self._apply_potential(lmp, elements, potential)
            lmp.run(0)
            S = lmp.thermo.computes['thermo_press'].vector
            results.append(MDReader(
                forces=lmp.system.forces.copy(),
                energy=lmp.thermo.computes['thermo_pe'].scalar + lmp.thermo.computes['my_ke'].scalar,
                stress=np.array([
                    [S[0], S[3], S[5]],
                    [S[3], S[1], S[4]],
                    [S[5], S[4], S[2]]
                ]),
                structure=structure))
        return results

    def shutdown(self):
        pass
